/*
 * condnet - LD_PRELOAD wrapper for per-process firewall kinda thing
 *
 * Copyright (C) 2012-2015 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses>.
 */


#define _GNU_SOURCE

#include <dlfcn.h>	/* dlsym */
#include <sys/types.h>	/* connect, getaddrinfo, sendto */
#include <sys/socket.h>	/* connect, getaddrinfo, sendto */
#include <netdb.h>	/* getaddrinfo */
#include <stdio.h>	/* *printf */
#include <errno.h>	/* errno */
#include <string.h>	/* memcpy */
#include <stdlib.h>	/* free */

#include "internal.h"
#include "log.h"
#include "filter.h"
#include "format.h"
#include "config.h"

#define NOFUNC_ERR	ENOMEM


char *wrapped_functions[] = {
	"connect",
	"getaddrinfo",
	"sendto",
	"accept",
	NULL,
};




static void
log_match_alter(const char *function, const char *orig_object,
		const char *spoof_object,
		struct action *action, int match)
{
	if (strcmp(orig_object, spoof_object) == 0) {
		if (action->set_ret) {
			char ret[32];
			action_to_str(ret, sizeof(ret), action);
			log_info("%s(%s) %s ALTER %s",
					function,
					orig_object,
					match ? "rule" : "policy",
					ret);
		} else {
			log_info("%s(%s) ACCEPT", function, orig_object);
		}
	} else {
		if (action->set_ret) {
			char ret[32];
			action_to_str(ret, sizeof(ret), action);
			log_info("%s(%s -> %s) %s ALTER %s",
					function, orig_object, spoof_object,
					match ? "rule" : "policy",
					ret);
		} else {
			log_info("%s(%s -> %s) ACCEPT",
					function, orig_object, spoof_object);
		}
	}
}



int
connect(int sockfd, const struct sockaddr *addr, socklen_t addrlen)
{
	typedef int (ofunc)(int sockfd, const struct sockaddr *addr,
			socklen_t addrlen);
	static ofunc *func = NULL;

	struct sockaddr real_addr;
	socklen_t real_addrlen = 14;
	struct action action;
	int match;
	int func_retval;

	config_setup();

	if (! func) {
		func = dlsym(RTLD_NEXT, "connect");
	}

	if (! func) {
		log_error("connect: symbol not found");
		errno = NOFUNC_ERR;
		return -1;
	}

	if (internal_is()) {
		log_debug("connect() internal");
		return func(sockfd, addr, addrlen);
	}
	log_debug("connect() managed");

	match = filter_outbound(addr, addrlen, &real_addr, &real_addrlen,
			&action);

	{
		char wanted_addr[ADDR_LEN] = "-";
		char got_addr[ADDR_LEN] = "-";
		sockaddr_to_str(wanted_addr, ADDR_LEN, addr, addrlen);
		sockaddr_to_str(got_addr, ADDR_LEN, &real_addr, real_addrlen);
		log_match_alter("connect", wanted_addr, got_addr,
				&action, match);
	}

	if (! action.set_ret) {
		internal_set();
		func_retval = func(sockfd, &real_addr, real_addrlen);
		internal_unset();
	} else {
		if (action.set_err) {
			errno = action.err;
		}
		func_retval = action.ret;
	}

	return func_retval;
}



int
getaddrinfo(const char *node, const char *service,
		const struct addrinfo *hints,
		struct addrinfo **res)
{
	typedef int (ofunc)(const char *node, const char *service,
		const struct addrinfo *hints,
		struct addrinfo **res);
	static ofunc *func = NULL;

	char *real_node = NULL;
	char *real_service = NULL;
	struct action action;
	int match;
	int func_retval;

	config_setup();

	if (! func) {
		func = dlsym(RTLD_NEXT, "getaddrinfo");
	}

	if (! func) {
		log_error("getaddrinfo: symbol not found");
		errno = NOFUNC_ERR;
		return -1;
	}

	if (internal_is()) {
		log_debug("getaddrinfo() internal");
		return func(node, service, hints, res);
	}
	log_debug("getaddrinfo() managed");

	match = filter_resolve(node, service, &real_node, &real_service,
			&action);
	{
		char wanted[128];
		char got[128];
		snprintf(wanted, sizeof(wanted), "%s/%s", node, service);
		snprintf(got, sizeof(got), "%s/%s", real_node, real_service);
		wanted[sizeof(wanted) - 1] = got[sizeof(got) - 1] = 0;
		log_match_alter("resolve", wanted, got, &action, match);
	}

	if (! action.set_ret) {
		internal_set();
		func_retval = func(real_node, real_service, hints, res);
		internal_unset();
	} else {
		if (action.set_err) {
			errno = action.err;
		}
		func_retval = action.ret;
	}

	free(real_node);
	free(real_service);

	return func_retval;
}



ssize_t
sendto(int sockfd, const void *buf, size_t len, int flags,
		const struct sockaddr *dest_addr, socklen_t addrlen)
{
	typedef int (ofunc)(int sockfd, const void *buf, size_t len, int flags,
			const struct sockaddr *dest_addr, socklen_t addrlen);
	static ofunc *func = NULL;

	struct sockaddr real_addr;
	socklen_t real_addrlen = 14;
	struct action action;
	int match;
	int func_retval;

	config_setup();

	if (! func) {
		func = dlsym(RTLD_NEXT, "sendto");
	}

	if (! func) {
		log_error("sendto: symbol not found");
		errno = NOFUNC_ERR;
		return -1;
	}

	if (internal_is()) {
		log_debug("sendto() internal");
		return func(sockfd, buf, len, flags, dest_addr, addrlen);
	}
	log_debug("sendto() managed");

	match = filter_outbound(dest_addr, addrlen, &real_addr, &real_addrlen,
			&action);

	{
		char wanted_addr[ADDR_LEN] = "-";
		char got_addr[ADDR_LEN] = "-";
		sockaddr_to_str(wanted_addr, ADDR_LEN, dest_addr, addrlen);
		sockaddr_to_str(got_addr, ADDR_LEN, &real_addr, real_addrlen);
		log_match_alter("sendto", wanted_addr, got_addr,
				&action, match);
	}

	if (! action.set_ret) {
		internal_set();
		func_retval = func(sockfd, buf, len, flags, &real_addr, real_addrlen);
		internal_unset();
	} else {
		if (action.set_err) {
			errno = action.err;
		}
		func_retval = action.ret;
	}


	return func_retval;
}



int
accept(int sockfd, struct sockaddr *addr, socklen_t *addrlen)
{
	typedef int (ofunc)(int sockfd, struct sockaddr *addr,
			socklen_t *addrlen);
	static ofunc *func = NULL;

	socklen_t pass_addrlen;
	struct sockaddr real_addr;
	socklen_t real_addrlen = 14;
	struct action action;
	int match;
	int func_retval;

	config_setup();

	if (! func) {
		func = dlsym(RTLD_NEXT, "accept");
	}

	if (! func) {
		log_error("accept: symbol not found");
		errno = NOFUNC_ERR;
		return -1;
	}

	if (internal_is()) {
		log_debug("accept() internal");
		func_retval = func(sockfd, addr, addrlen);
		return func_retval;
	}
	log_debug("accept() managed");

	internal_set();
	func_retval = func(sockfd, addr, addrlen);
	internal_unset();

	pass_addrlen = addrlen ? *addrlen : 0;
	match = filter_inbound(addr, pass_addrlen, &real_addr, &real_addrlen,
			&action);

	{
		char wanted_addr[ADDR_LEN] = "-";
		char got_addr[ADDR_LEN] = "-";
		sockaddr_to_str(wanted_addr, ADDR_LEN, addr, pass_addrlen);
		sockaddr_to_str(got_addr, ADDR_LEN, &real_addr, real_addrlen);
		log_match_alter("accept", wanted_addr, got_addr,
				&action, match);
	}

	if (! action.set_ret) {
		/* nothing here */
	} else {
		if (action.set_err) {
			errno = action.set_err;
		}
		func_retval = action.ret;
	}

	return func_retval;
}



#if 0
int
recvfrom(int sockfd, void *buf, size_t len, int flags,
		struct sockaddr *src_addr, socklen_t *addrlen)
{
}



int
bind(int sockfd, const struct sockaddr *addr, socklen_t addrlen)
{
}
#endif
