#include "config.h"
#include "rules.h"
#include "parser.h"
#include "log.h"

#include <stdlib.h>	/* malloc, calloc, getenv, atexit */
#include <assert.h>	/* assert */
#include <errno.h>	/* errno */
#include <string.h>	/* strerror, strchr, memmove */
#include <stdint.h>	/* uint16_t */
#include <stdio.h>	/* *printf, fopen, fclose */
#include <sys/types.h>	/* open */
#include <sys/stat.h>	/* open */
#include <fcntl.h>	/* open */



#define RC_CHUNK	4096



struct rules *condnet_config = NULL;
static int config_done = 0;



static char *
config_filename(void)
{
	char *file;
	file = getenv(ENV_FILE);
	if (file == NULL) {
		const char *home;
		size_t len;
		home = getenv("HOME");
		if (home == NULL) {
			return NULL;
		}
		len = strlen(home) + 1 + strlen(CONFIG_FILE) + 1;
		file = malloc(len);
		assert(file);

		snprintf(file, len, "%s/" CONFIG_FILE, home);
		file[len - 1] = '\0';
	} else {
		file = strdup(file);
	}
	return file;
}



struct rules *
config_from_file(const char *filename)
{
	struct rules *rules;

	FILE *fptr;
	char *buf;

	char *line_start, *line_end;

	size_t buf_used;
	size_t read_bytes;
	size_t want_bytes;

	if (filename == NULL) {
		return NULL;
	}

	buf = malloc(RC_CHUNK + 1);
	assert(buf);

	fptr = fopen(filename, "r");
	if (fptr == NULL) {
		free(buf);
		log_warn("%s: cannot read configuration", filename);
		return NULL;
	}
	log_debug("reading configuration from file \"%s\"", filename);

	rules = rules_new();
	assert(rules);

	buf_used = 0;
	while (1) {
		want_bytes = RC_CHUNK - buf_used;
		if (want_bytes == 0) {
			log_warn("%s: line too long (> %d bytes), aborting",
					filename, RC_CHUNK);
			buf_used = 0; /* don't parse remaining buffer */
			break;
		}

		read_bytes = fread(buf + buf_used, 1, want_bytes, fptr);
		if (read_bytes == 0) {
			if (feof(fptr)) {
				break;
			}
		}
		buf_used += read_bytes;
		buf[buf_used] = '\0';

		line_start = buf;
		line_end = strchr(line_start, '\n');
		while (line_end) {
			*line_end = '\0';
			parser_add_line(rules, line_start);
			line_start = line_end + 1;
			if (line_start < buf + buf_used) {
				line_end = strchr(line_start, '\n');
			} else {
				line_end = NULL;
			}
		}
		if (line_start == buf + buf_used) {
			buf_used = 0;
		} else {
			memmove(buf, line_start, buf + buf_used - line_start);
			buf_used -= line_start - buf;
		}
	}
	if (buf_used) {
		parser_add_line(rules, buf);
	}

	fclose(fptr);
	free(buf);

	return rules;
}



struct rules *
config_from_env(void)
{
	char *env;
	env = getenv(ENV_CONFIG);

	if (env == NULL) {
		return NULL;
	} else {
		log_debug("reading configuration from environment");
		return parser_buffer_to_rules(env);
	}
}



static void
free_config(void)
{
	log_debug("free configuration");
	if (condnet_config) {
		rules_free(condnet_config);
	}
}



static void
configure(void)
{
	condnet_config = config_from_env();
	if (condnet_config == NULL) {
		char *filename = config_filename();
		condnet_config = config_from_file(filename);
		free(filename);
	}
	if (condnet_config == NULL) {
		condnet_config = rules_new();
	}
	rules_to_log(condnet_config);
	atexit(free_config);
}



inline void
config_setup(void)
{
	if (config_done) {
		return;
	}
	configure();
	config_done = 1;
}



int
config_rule_to_str(char *buf, size_t len, const struct connection_rule *rule)
{
	return snprintf(buf, len, "**%s/%d%s -> %s/%d%s (= %s/%d) -> %d%s/%d%s**",
			rule->from_addr ? : "*", rule->from_port,
			rule->from_rev ? "/reverse" : "",
			rule->to_addr ? : "*", rule->to_port,
			rule->to_rev ? "/reverse" : "",
			rule->spoof_addr ? : "-", rule->spoof_port,
			rule->policy.ret, rule->policy.set_ret ? "!" : "~",
			rule->policy.err, rule->policy.set_err ? "!" : "~");
}
