#include "rules.h"
#include "log.h"

#include <string.h>	/* strlen, memcpy */
#include <stdlib.h>	/* *alloc */
#include <assert.h>	/* assert */



char *	/* XXX static */
strdup_or_null(const char *str)
{
	char *dup;
	size_t len;

	if (str == NULL) {
		return NULL;
	}

	len = strlen(str);
	dup = malloc(strlen(str) + 1);
	assert(dup);
	strncpy(dup, str, len);
	dup[len] = 0;

	return dup;
}



static void
free_resolve(struct resolve_rule *rule)
{
	free(rule->node_pattern);
	free(rule->service_pattern);
	free(rule->spoof_node);
	free(rule->spoof_service);
	free(rule);
}



static void
free_connection(struct connection_rule *rule)
{
	free(rule->from_addr);
	free(rule->to_addr);
	free(rule->spoof_addr);
	free(rule);
}



void
rules_free(struct rules *rules)
{
	int i;

	if (rules == NULL) {
		return;
	}

	for (i = 0; i < rules->resolve_n; i++) {
		struct resolve_rule *rule;
		rule = rules->resolve[i];
		free_resolve(rule);
	}
	free(rules->resolve);
	for (i = 0; i < rules->in_n; i++) {
		struct connection_rule *rule;
		rule = rules->in[i];
		free_connection(rule);
	}
	free(rules->in);
	for (i = 0; i < rules->out_n; i++) {
		struct connection_rule *rule;
		rule = rules->out[i];
		free_connection(rule);
	}
	free(rules->out);
	free(rules);
}



struct rules *
rules_new(void)
{
	struct rules *rules;
	rules = malloc(sizeof(struct rules));
	assert(rules);

	action_zero(&(rules->in_policy));
	action_zero(&(rules->out_policy));
	action_zero(&(rules->resolve_policy));

	rules->resolve = malloc(sizeof(struct resolve_rule *));
	assert(rules->resolve);
	rules->resolve[0] = NULL;
	rules->resolve_n = 0;

	rules->in = malloc(sizeof(struct connection_rule *));
	assert(rules->in);
	rules->in[0] = NULL;
	rules->in_n = 0;

	rules->out = malloc(sizeof(struct connection_rule *));
	assert(rules->out);
	rules->out[0] = NULL;
	rules->out_n = 0;

	return rules;
}



void
rules_add_inbound(struct rules *rules, struct connection_rule *rule)
{
	rules->in = realloc(rules->in, sizeof(struct connection_rule *) *
				(rules->in_n + 2));
	assert(rules->in);
	rules->in[rules->in_n] = rule;
	rules->in[rules->in_n + 1] = NULL;
	rules->in_n++;
}



void
rules_add_outbound(struct rules *rules, struct connection_rule *rule)
{
	rules->out = realloc(rules->out, sizeof(struct connection_rule *) *
				(rules->out_n + 2));
	assert(rules->out);
	rules->out[rules->out_n] = rule;
	rules->out[rules->out_n + 1] = NULL;
	rules->out_n++;
}



void
rules_add_resolve(struct rules *rules, struct resolve_rule *rule)
{
	rules->resolve = realloc(rules->resolve,
			sizeof(struct resolve_rule *) * (rules->resolve_n + 2));
	assert(rules->resolve);
	rules->resolve[rules->resolve_n] = rule;
	rules->resolve[rules->resolve_n + 1] = NULL;
	rules->resolve_n++;
}



struct resolve_rule *
rule_resolve_new(const char *node_pattern, const char *service_pattern,
		const char *spoof_node, const char *spoof_service,
		struct action *action)
{
	struct resolve_rule *rule;

	rule = malloc(sizeof(struct resolve_rule));
	assert(rule);

	rule->node_pattern = strdup_or_null(node_pattern);
	rule->service_pattern = strdup_or_null(service_pattern);
	rule->spoof_node = strdup_or_null(spoof_node);
	rule->spoof_service = strdup_or_null(spoof_service);

	action_copy(&(rule->policy), action);

	return rule;
}



struct connection_rule *
rule_connection_new(int protocol, /* FIXME action */
		const char *from_addr, uint16_t from_port, int from_rev,
		const char *to_addr, uint16_t to_port, int to_rev,
		const char *spoof_addr, uint16_t spoof_port,
		struct action *action)
{
	struct connection_rule *rule;

	rule = malloc(sizeof(struct connection_rule));
	assert(rule);

	action_copy(&(rule->policy), action);

	rule->protocol = protocol;
	rule->to_addr = strdup_or_null(to_addr);
	rule->to_port = to_port;
	rule->to_rev = to_rev;
	rule->from_addr = strdup_or_null(from_addr);
	rule->from_port = from_port;
	rule->from_rev = from_rev;

	rule->spoof_addr = strdup_or_null(spoof_addr);
	rule->spoof_port = spoof_port;

	return rule;
}



static void
connection_to_stdout(struct connection_rule *rule)
{
	char policy[32];
	action_to_str(policy, sizeof(policy), &(rule->policy));
	log_debug("  proto=%d from=%s/%d%s to=%s/%d%s spoof=%s/%d -> %s",
			rule->protocol,
			rule->from_addr ? : "-",
			rule->from_port ? : 0,
			rule->from_rev ? "/reverse" : "",
			rule->to_addr ? : "-",
			rule->to_port ? : 0,
			rule->to_rev ? "/reverse" : "",
			rule->spoof_addr ? : "-",
			rule->spoof_port ? : 0,
			policy);
}



static void
resolve_to_stdout(struct resolve_rule *rule)
{
	char policy[32];
	action_to_str(policy, sizeof(policy), &(rule->policy));
	log_debug("  pattern=%s/%s spoof=%s/%s -> %s",
			rule->node_pattern ? : "-",
			rule->service_pattern ? : "-",
			rule->spoof_node ? : "-",
			rule->spoof_service ? : "-",
			policy);
}



void
rules_to_log(struct rules *rules)
{
	struct connection_rule **conn;
	struct resolve_rule **resolve;
	char policy[32];

	if (rules == NULL) {
		log_debug("NO RULES");
		return;
	}

	action_to_str(policy, sizeof(policy), &(rules->in_policy));
	log_debug("INPUT (policy=%s)", policy);
	for (conn = rules->in; *conn; conn++) {
		connection_to_stdout(*conn);
	}
	action_to_str(policy, sizeof(policy), &(rules->out_policy));
	log_debug("OUTPUT (policy=%s)", policy);
	for (conn = rules->out; *conn; conn++) {
		connection_to_stdout(*conn);
	}
	action_to_str(policy, sizeof(policy), &(rules->resolve_policy));
	log_debug("RESOLVE (policy=%s)", policy);
	for (resolve = rules->resolve; *resolve; resolve++) {
		resolve_to_stdout(*resolve);
	}
	log_debug("");
}



void
action_set(struct action *action, int ret, int err)
{
	action->ret = ret;
	action->set_ret = ret ? 1 : 0;
	action->err = err;
	action->set_err = ret ? 1 : 0;
}



inline void
action_copy(struct action *dest, struct action *src)
{
	memcpy(dest, src, sizeof(struct action));
}


inline void
action_zero(struct action *action)
{
	action->ret = action->set_ret = 0;
	action->err = action->set_err = 0;
}



void
action_to_str(char *str, size_t len, struct action *action)
{
	char ret[16], err[16];
	if (action->set_ret) {
		snprintf(ret, sizeof(ret) - 1, "%d", action->ret);
	} else {
		ret[0] = '-';
		ret[1] = 0;
	}
	if (action->set_err) {
		snprintf(err, sizeof(ret) - 1, "%d", action->err);
	} else {
		err[0] = '-';
		err[1] = 0;
	}
	snprintf(str, len, "%s/%s", ret, err);
	str[len - 1] = 0;
}
